/* 
 *
 * This file is part of the 'NewsService Journaline(R) Decoder'
 * 
 * Copyright (c) 2003, 2004 by Fraunhofer IIS, Erlangen, Germany
 * 
 * --------------------------------------------------------------------
 * 
 * For NON-COMMERCIAL USE,
 * the 'NewsService Journaline(R) Decoder' is free software;
 * you can redistribute it and/or modify it under the terms of
 * the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * The 'NewsService Journaline(R) Decoder' is distributed in the hope
 * that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with the 'NewsService Journaline(R) Decoder';
 * if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 * 
 * If you use this software in a project with user interaction, please
 * provide the following text to the user in an appropriate place:
 * "Features NewsService Journaline(R) decoder technology by
 * Fraunhofer IIS, Erlangen, Germany.
 * For more information visit http://www.iis.fhg.de/dab"
 * 
 * --------------------------------------------------------------------
 * 
 * To use the 'NewsService Journaline(R) Decoder' software for
 * COMMERCIAL purposes, please contact Fraunhofer IIS for a
 * commercial license (see below for contact information)!
 * 
 * --------------------------------------------------------------------
 * 
 * Contact:
 *   Fraunhofer IIS, Department 'Broadcast Applications'
 *   Am Wolfsmantel 33, 91058 Erlangen, Germany
 *   http://www.iis.fraunhofer.de/dab
 *   mailto:bc-info@iis.fraunhofer.de
 * 
 */

//////////////////////////////////////////////////////////////////////////////
/// Journaline(R) news service decoder demonstration program
///
/// @file      demo.cpp
/// @author    Michael Reichenbaecher <rbr@iis.fraunhofer.de>
///
/// $Id: demo.cpp,v 1.1 2004/03/03 11:04:30 rbr Exp $
///
/// Module:    Journaline(R)
///
/// Copyright: (C) 2004 Fraunhofer Institut Integrierte Schaltungen, Erlangen
///
/// Compiler:  gcc version 3.3 20030226 (prerelease)
///            Microsoft Visual C++ .NET
///
/// This demonstration program creates a DAB data group decoder
/// and a Journaline(R) news service decoder and feeds the files
/// specified on the command line to the data group decoder.
///
/// The user can interactively add further data group files and
/// display arbitrary NML news objects by specifying their object id
///
/// @note
///   In a real application, there should be a service manager handling
///   (possibly) multiple instances of the two decoder types.
///   In this demonstration, we only use one global instance of each.
//////////////////////////////////////////////////////////////////////////////


#ifdef LINUX
#include <stdio.h>
#endif

#ifdef HAVE_READLINE
#include <readline/readline.h>
#include <readline/history.h>
#endif

#ifdef WIN32
#pragma warning(push,3)
#endif

#include <string>
#include <iostream>

#ifdef WIN32
#pragma warning(pop)
#endif

#include "NML.h"
#include "newssvcdec.h"
#include "dabdatagroupdecoder.h"


/// a unique instance of a DAB data group decoder
static DAB_DATAGROUP_DECODER_t dgdec = NULL;

/// a unique instance of a Journaline(R) news service decoder
static NEWS_SVC_DEC_decoder_t newsdec = NULL;


// set these to 1 for more debug information
#ifdef WIN32
extern "C" int showDdNewsSvcDecErr = 0;
extern "C" int showDdNewsSvcDecInfo = 0;
extern "C" int showDdDabDgDecErr = 0;
extern "C" int showDdDabDgDecInfo = 0;
#else
int showDdNewsSvcDecErr = 0;
int showDdNewsSvcDecInfo = 0;
int showDdDabDgDecErr = 0;
int showDdDabDgDecInfo = 0;
#endif


/*!
 * @brief read data group from file
 *
 * reads a DAB data group from file and feeds
 * it to the DAB data group decoder
 * 
 * @param filename  name of file
 * @retval 0  failure
 * @retval 1  success
 */
int ReadDatagroupFile(const char *filename)
{
  unsigned char buf[2048];
  unsigned long buf_len;

  FILE *f = fopen(filename, "rb");
  if (!f)
  {
    perror("fopen");
    return 0;
  }
  buf_len = fread(buf, 1, 2048, f);
  if (ferror(f))
  {
    perror("fread");
    fclose(f);
    return 0;
  }
  fclose(f);

  if (!DAB_DATAGROUP_DECODER_putData(dgdec, buf_len, buf))
  {
    fprintf(stderr, "DAB_DATAGROUP_DECODER_putData: error");
    return 0;
  }

  return 1;
}



/*!
 * @brief callback function for data group decoder
 * 
 * Will be called when a valid DAB datagroup has been evaluated
 * by the DAB data group decoder.
 * Feeds the data field to the Journaline(R) news service decoder
 */
void dg_cb(const DAB_DATAGROUP_DECODER_msc_datagroup_header_t * /*header*/,
           const unsigned long len,
           const unsigned char *buf,
           void * /*arg*/ )
{
  //printf("dg_cb: NML =\n%s\n", HexDump(buf, len).c_str());
  if (!NEWS_SVC_DEC_putData(newsdec, len, buf))
  {
    fprintf(stderr, "dg_cb: NEWS_SVC_DEC_putData failed!");
  }
}



/*!
 * @brief callback function for Journaline(R) news decoder
 *
 * will be called when the availability of watched news objects
 * changes
 */
void obj_avail_cb(unsigned long number_of_elements,
                  NEWS_SVC_DEC_obj_availability_t *chg_list,
                  void * /*arg*/ )
{
  printf("obj_avail_cb\n");
  for (unsigned long i=0; i<number_of_elements; i++)
  {
    printf("object #%04x now has status %d\n", chg_list[i].object_id, chg_list[i].status);
  }
}


/*!
 * @brief select current object
 * @param oid  news object id
 * @retval 0  failure
 */
int SelectObject(NML::NewsObjectId_t oid)
{
  NML::RawNewsObject_t rno;
  unsigned long elen = 0;
  unsigned long len = 0;
  if (!NEWS_SVC_DEC_get_news_object(newsdec,
                                    oid,
                                    &elen,
                                    &len,
                                    rno.nml))
  {
    fprintf(stderr, "news object #%04x not available\n", oid);
    return 0;
  }
  rno.nml_len = static_cast<unsigned short>(len);
  rno.extended_header_len = static_cast<unsigned short>(elen);
  printf("%s\n", DumpRaw(rno).c_str());

  RemoveNMLEscapeSequences handler;
  NML *nml = NMLFactory::Instance()->CreateNML(rno, &handler);


  // display NML
  printf("Object #%04x:\n", oid);
  printf("-------------\n");
  printf("%s\n", nml->Dump().c_str());

  //TODO: set watches, keeps, ...
  
  delete nml;
  return 1;
}


/*!
 * @brief command help
 *
 * in-program command help
 */
void cmd_help(void)
{
  printf("\nAvailable Commands are:\n");
  printf("o <hexadecimal object_id> => show object (o 0000 for root object)\n");
  printf("r <filename>              => feed file to data group decoder\n");
  printf("h                         => this help\n");
  printf("q                         => quit\n");
}


/*!
 * @brief read a user command
 *
 * reads one line of user input (max. 80 characters)
 *
 * @return command line
 */
std::string cmd_read(void)
{
#ifndef HAVE_READLINE
  char cmd[80];
  printf("cmd> ");
  fgets(cmd, 80, stdin);
  cmd[strlen(cmd)-1] = 0;
  return std::string(cmd);
#else
  std::string s;
  char *cmd= readline("cmd> ");
  s = cmd?cmd:"";
  return s;
#endif
}




/*!
 * @brief news service decoder demonstration main program
 * @param argc number of arguments + 1
 * @param argv arguments
 * @return exit status
 */
int main(int argc, char *argv[])
{
  printf("Journaline(R) news service decoder demonstration program\n");
  printf("Copyright (C)2004 Fraunhofer Institute Integrated Circuits IIS-A\n");

  // -----------------------------------------------------------------------
  // create dab data group decoder
  // -----------------------------------------------------------------------
  dgdec = DAB_DATAGROUP_DECODER_createDec(dg_cb, NULL);
  if (!dgdec)
  {
    fprintf(stderr, "DAB_DATAGROUP_DECODER_createDec failed\n");
    return 1;
  }

  // -----------------------------------------------------------------------
  // create Journaline(R) news decoder
  // -----------------------------------------------------------------------
  // 1 MB memory for news decoder
  unsigned long max_memory = 1024*1024;

  // No limit for number of NML objects (except max_memory limitation)
  unsigned long max_objects = 0;

  // no extended header will be used in this demonstration
  unsigned long extended_header_len = 0;
  newsdec = NEWS_SVC_DEC_createDec(obj_avail_cb,
                                   max_memory,
                                   &max_objects,
                                   extended_header_len,
                                   NULL);
  if (!newsdec)
  {
    fprintf(stderr, "NEWS_SVC_DEC_createDec failed\n");
    return 2;
  }

  // -----------------------------------------------------------------------
  // read files specified on command line
  // -----------------------------------------------------------------------
  int i;
  for (i=1; i<argc; i++)
  {
    if (!ReadDatagroupFile(argv[i]))
    {
      fprintf(stderr, "could not read dg");
      continue;
    }
    printf("Read datagroup from file #%d '%s'\n", i, argv[i]);
  }


  // -----------------------------------------------------------------------
  // interactive section
  // -----------------------------------------------------------------------
  std::string cmd;

  cmd_help();
  for(;;)
  {
    cmd = cmd_read();

    if ((cmd=="") || (cmd[0]=='q')) {break;}
    switch(cmd[0])
    {
    case '\0':
      break;
    case 'o':
      if (cmd[1]!=' ') break;
      unsigned short oid;
      if (sscanf(cmd.c_str() + 2, "%04hx", &oid)!=1)
      {
        fprintf(stderr,"wrong oid\n");
      }
      else
      {
        SelectObject(oid);
      }
      break;
    case 'r':
      if (cmd[1]!=' ') break;
      ReadDatagroupFile(cmd.c_str()+2);
      break;
    default:
      cmd_help();
    }
#ifdef HAVE_READLINE
    add_history(cmd.c_str());
#endif
  }


  // -----------------------------------------------------------------------
  // cleanup
  // -----------------------------------------------------------------------
  NEWS_SVC_DEC_deleteDec(newsdec);
  DAB_DATAGROUP_DECODER_deleteDec(dgdec);

  return 0;
}
